#include <stdio.h>

#include "wmsstats.h"
#include "wmsclass.h"

#define NANOSECSINSEC ((long) 1000000000)

ClassStatsRec *methods = (ClassStatsRec *) NULL;

void print_method_stats ();
ClassStatsRec *find_method_stat_rec ();

void usage ()
{
   printf ("usage: wmsclass [-f filename] -c classname\n");
   exit (1);
}

main (argc, argv)
   int argc;
   char *argv[];
{
   register int i;
   char *arg;
   char *widget_class = (char *) NULL;
   char *default_file_name = "wms.data";
   char *file_name = default_file_name;
   FILE *fp;
   WidgetStatisticsDataRec datarec;
   ClassStatsRec *method_stat_rec;

   /* parse the command line arguments */
   for (i = 1; i < argc; i++)
   {
      arg = argv[i];

      if (arg[0] == '-')
      {
         switch (arg[1])
         {
            /* the widget class name */
            case 'c':
               if (++i >= argc)
                  usage ();
               widget_class = argv[i];
               break;

            /* the data file name */
            case 'f':
               if (++i >= argc)
                  usage ();
               file_name = argv[i];
               break;

            default:
               usage ();
         }
      }
   }

   if (widget_class == (char *) NULL)
      usage ();

   /* open file */
   if ((fp = fopen (file_name, "r")) == (FILE *) NULL)
   {
      printf ("wmsclass: cannot open data file %s\n", file_name);
      exit (1);
   }

   /* read the statistics file one record at a time */
   while (fread (&datarec, sizeof (datarec), 1, fp) == 1)
   {
      if (strcmp (datarec.method_class, widget_class) == 0)
      {
         method_stat_rec = find_method_stat_rec (datarec.method_string);

         /* increment the total number of times the method was called */
         (method_stat_rec->times_called)++;

         /* increment the time spent in the method */
         method_stat_rec->secs += datarec.secs;
         if (datarec.nanosecs + method_stat_rec->nanosecs >= NANOSECSINSEC)
         {
            method_stat_rec->secs++;
            method_stat_rec->nanosecs = datarec.nanosecs + method_stat_rec->nanosecs
               - NANOSECSINSEC;
         }
         else
            method_stat_rec->nanosecs += datarec.nanosecs;
      }
   }

   /* output the statistics */
   printf ("\n\t%20s: %s\n", "Method Class", widget_class);
   printf ("\t%20s: %s\n\t%20s: %s\n\t%20s: \n\n", "Produced By", "wmsclass", 
      "File Name", file_name, "Date/Time");
   printf ("%-32s   %-12s   %-4s   %8s\n", "Method", "Times Called", "Secs", "Nanosecs");
   printf ("%-32s   %-12s   %-4s   %8s\n", "------", "------------", "----", "--------");

   print_method_stats ("class_initialize");
   print_method_stats ("class_part_initialize");
   print_method_stats ("initialize");
   print_method_stats ("initialize_hook");
   print_method_stats ("realize");
   print_method_stats ("destroy");
   print_method_stats ("resize");
   print_method_stats ("expose");
   print_method_stats ("set_values");
   print_method_stats ("set_values_hook");
   print_method_stats ("set_values_almost");
   print_method_stats ("get_values_hook");
   print_method_stats ("accept_focus");
   print_method_stats ("query_geometry");
   print_method_stats ("display_accelerator");
}

/* Utility routine to print out statistics for a method */
void print_method_stats (method_name)
   char *method_name;
{
   ClassStatsRec *method_stat_rec;

   method_stat_rec = find_method_stat_rec (method_name);

   printf ("%-32s   %12d   %4d  %9d\n",
      method_stat_rec->method_name, method_stat_rec->times_called, 
      method_stat_rec->secs, method_stat_rec->nanosecs);
}

/* utility routine to find a data record for a method */
ClassStatsRec *find_method_stat_rec (method_name)
   char *method_name;
{
   ClassStatsRec *ptr = methods;

   /* search for the method in the list of the records */
   while (ptr != NULL)
   {
      if (strcmp (ptr->method_name, method_name) == 0)
         return (ptr);
      ptr = ptr->next;
   }

   /* allocate a new record */
   ptr = (ClassStatsRec *) XtCalloc (1, sizeof (ClassStatsRec));

   /* initialize the fields of the record */
   strcpy (ptr->method_name, method_name);
   ptr->times_called = 0;
   ptr->secs = 0L;
   ptr->nanosecs = 0L;

   /* add the record to the head of the list */
   ptr->next = methods;
   methods = ptr;

   return (ptr);
}
