#include <stdio.h>

#include "wmsstats.h"
#include "wmsmethod.h"

#define True  1
#define False 0

MethodStatsRec *widgets = (MethodStatsRec *) NULL;
MethodStatsRec *sorted = (MethodStatsRec *) NULL;

void usage ()
{
   printf ("usage: wmsmethod [-f filename] -c classname -m methodname\n");
   exit (1);
}

#define NANOSECSINSEC ((long) 1000000000)

main (argc, argv)
   int argc;
   char *argv[];
{
   register int i;
   char *arg;
   char *method_class = (char *) NULL;
   char *method_name = (char *) NULL;
   char *default_file_name = "wms.data";
   char *file_name = default_file_name;
   FILE *fp;
   WidgetStatisticsDataRec datarec;
   MethodStatsRec *widget_stat_rec;
   MethodStatsRec *find_widget_stat_rec ();
   MethodStatsRec *sorted_list;
   MethodStatsRec *sort_by_times_called ();
   MethodStatsRec *ptr = sorted_list;
   long total_times_called = 0;
   long total_secs = 0;
   long total_nanosecs = 0;
   short is_a_class_method = False;

   /* parse the command line arguments */
   for (i = 1; i < argc; i++)
   {
      arg = argv[i];

      if (arg[0] == '-')
      {
         switch (arg[1])
         {
            /* the widget class name */
            case 'c':
               if (++i >= argc)
                  usage ();
               method_class = argv[i];
               break;

            /* the data file name */
            case 'f':
               if (++i >= argc)
                  usage ();
               file_name = argv[i];
               break;

            /* the method name */
            case 'm':
               if (++i >= argc)
                  usage ();
               method_name = argv[i];
               break;
            default:
               usage ();
         }
      }
   }

   /* verify that a widget class and a method were both specified */
   if (method_class == (char *) NULL || method_name == (char *) NULL)
      usage ();

   /* open the statistics file */
   if ((fp = fopen (file_name, "r")) == (FILE *) NULL)
   {
      printf ("wmsmethod: cannot open data file %s\n", file_name);
      exit (1);
   }

   /* read the statistics file one record at a time */
   while (fread (&datarec, sizeof (datarec), 1, fp) == 1)
   {
      /* check to see if the record read in matches the method_class
         and method_name specified by the user */
      if (strcmp (datarec.method_string, method_name) == 0 &&
         strcmp (datarec.method_class, method_class) == 0)
      {
         /* find/get a record to tally the statistics for the widget */
         if (strlen (datarec.widget_name) == 0)
         {
            widget_stat_rec = find_widget_stat_rec (datarec.method_class, "");
            is_a_class_method = True;
         }
         else
            widget_stat_rec = find_widget_stat_rec (datarec.widget_name,
               datarec.widget_class);

         /* increment the total number of times the method was called */
         total_times_called++;

         /* increment the number of times the method was called for the
            specific widget */
         (widget_stat_rec->times_called)++;

         /* increment the total time spent in the method */
         total_secs += datarec.secs;
         if (datarec.nanosecs + total_nanosecs >= NANOSECSINSEC)
         { /* the number of nanoseconds exceeds one second */
            total_secs++;
            total_nanosecs = datarec.nanosecs + total_nanosecs - NANOSECSINSEC;
         }
         else
            total_nanosecs += datarec.nanosecs;

         /* increment the time spent in the method for the widget */
         widget_stat_rec->secs += datarec.secs;
         if (datarec.nanosecs + widget_stat_rec->nanosecs >= NANOSECSINSEC)
         { /* the number of nanoseconds exceeds one second */
            widget_stat_rec->secs++;
            widget_stat_rec->nanosecs = datarec.nanosecs + widget_stat_rec->nanosecs
               - NANOSECSINSEC;
         }
         else
            widget_stat_rec->nanosecs += datarec.nanosecs;
      }
   }

   /* sort the widget records */
   sorted_list = sort_by_times_called ();

   /* output the statistics */
   printf ("\n\t%20s: %s\n\t%20s: %s\n", "Method Name", method_name, "Widget Class",
      method_class);
   printf ("\t%20s: %s\n\t%20s: %s\n\t%20s: \n\n", "Produced By", "wmsmethod", 
      "File Name", file_name, "Date Run");

   printf ("\tTotal Number of times called: %d\n", total_times_called);
   printf ("\tTotal time in method: %ld sec %ld nanosecs\n\n", total_secs, total_nanosecs);

   /* print out column headers */
   if (!is_a_class_method)
   {
      printf ("%-26s   %-26s   %-12s   %-4s    %8s\n", "Widget", "Widget Class",
         "Times Called", "Secs", "Nanosecs");
      printf ("%-26s   %-26s   %-12s   %-4s    %8s\n", "------", "------------",
         "------------", "----", "--------");
   }

   ptr = sorted_list;
   if (is_a_class_method)
   {
      printf ("There are no per widget statistics for the %s method.\n\n",
         method_name);
   }
   else
   {
      while (ptr != NULL)
      {
         printf ("%-26s   %-26s   %12d   %4ld   %9ld\n", 
            ptr->widget_name, ptr->widget_class, 
            ptr->times_called, ptr->secs, ptr->nanosecs);
         ptr= ptr->next;
      }
   }
}

MethodStatsRec *find_widget_stat_rec (widget_name, widget_class)
   char *widget_name;
   char *widget_class;
{
   MethodStatsRec *ptr = widgets;

   /* search for the widget in the list of widget statistics records */
   while (ptr != NULL)
   {
      if (strcmp (ptr->widget_name, widget_name) == 0)
         return (ptr);
      ptr = ptr->next;
   }

   /* allocate a new widget stat rec */
   ptr = (MethodStatsRec *) XtCalloc (1, sizeof (MethodStatsRec));

   /* initialize the fields of the new record */
   strcpy (ptr->widget_name, widget_name);
   strcpy (ptr->widget_class, widget_class);
   ptr->times_called = 0;
   ptr->secs = 0;
   ptr->nanosecs = 0;

   /* add the new record to the head of the list */
   ptr->next = widgets;
   widgets = ptr;

   return (ptr);
}

/********* Sorting code **********/

void insert_before (rectoinsert, prevrec, nextrec)
   MethodStatsRec *rectoinsert;
   MethodStatsRec *prevrec;
   MethodStatsRec *nextrec;
{
   if (prevrec == (MethodStatsRec *) NULL)
   { /* add to the front of the list */
      rectoinsert->next = sorted;
      sorted = rectoinsert;
   }
   else
   {
      rectoinsert->next = nextrec;
      prevrec->next = rectoinsert;
   }
}

void insert_after (rectoinsert, prevrec)
   MethodStatsRec *rectoinsert;
   MethodStatsRec *prevrec;
{
   rectoinsert->next = prevrec->next;
   prevrec->next = rectoinsert;
}
   
MethodStatsRec *sort_by_times_called ()
{
   MethodStatsRec *nextwidgettosort = widgets;
   MethodStatsRec *nextrectosort = widgets;
   MethodStatsRec *start, *prev = (MethodStatsRec *) NULL;
   MethodStatsRec *next = (MethodStatsRec *) NULL;

   /* Put the first record at the beginning of the sorted list */
   sorted = nextrectosort;
   if (sorted == (MethodStatsRec *) NULL)
      return (sorted);
   if (sorted->next != (MethodStatsRec *) NULL)
   {
      nextwidgettosort = sorted->next;
      sorted->next = (MethodStatsRec *) NULL;
   }
   else
      return (sorted);

   do 
   {
      /* Remove the nextrectosort from the list of widgets */
      nextrectosort = nextwidgettosort;
      nextwidgettosort = nextwidgettosort->next;
      nextrectosort->next = NULL;
      
      /* Insert nextrectosort in the appropriate position by searching
         through the list of records already sorted. */
      prev = (MethodStatsRec *) NULL;
      start = sorted;
      while (start != (MethodStatsRec *) NULL)
      {
         if (nextrectosort->times_called < start->times_called)
         {
            insert_before (nextrectosort, prev, start);
            break;
         }
         else if (start->next == (MethodStatsRec *) NULL)
         {
            insert_after (nextrectosort, start);
            break;
         }
         else
         {
            prev = start;
            start = start->next;
         }
      }
   } while (nextwidgettosort != (MethodStatsRec *) NULL);

   return (sorted);
}

